from importlib import metadata as _metadata

__version__ = _metadata.version("argos3")

# Main Components
from .convolutional import *
from .datagram import *
from .detector import *
from .encoder import *
from .formatter import *
from .lowpassfilter import *
from .matchedfilter import *    
from .modulator import *
from .multiplexer import *
from .preamble import *
from .sampler import * 
from .scrambler import *
from .synchronizer import *

# Channel
from .channel import *
from .noise import *

# Transmitter
from .transmitter import *

# Receiver
from .receiver import *

# Extras
from .bersnr import *
from .data import * 
from .plotter import *
